package envoy.client.ui.controller;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.paint.Color;

import envoy.client.data.Chat;
import envoy.client.data.LocalDB;
import envoy.client.data.Settings;
import envoy.client.event.MessageCreationEvent;
import envoy.client.net.Client;
import envoy.client.net.WriteProxy;
import envoy.client.ui.MessageListCell;
import envoy.client.ui.SceneContext;
import envoy.client.ui.ContactListCell;
import envoy.data.Contact;
import envoy.data.Message;
import envoy.data.MessageBuilder;
import envoy.event.EventBus;
import envoy.event.MessageStatusChangeEvent;
import envoy.event.UserStatusChangeEvent;
import envoy.event.contact.ContactOperationEvent;
import envoy.util.EnvoyLog;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>ChatSceneController.java</strong><br>
 * Created: <strong>26.03.2020</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @since Envoy Client v0.1-beta
 */
public final class ChatScene {

	@FXML
	private Label contactLabel;

	@FXML
	private ListView<Message> messageList;

	@FXML
	private ListView<Contact> userList;

	@FXML
	private Button postButton;

	@FXML
	private Button settingsButton;

	@FXML
	private TextArea messageTextArea;

	@FXML
	private Label remainingChars;

	private LocalDB			localDB;
	private Client			client;
	private WriteProxy		writeProxy;
	private SceneContext	sceneContext;

	private Chat currentChat;

	private static final Settings	settings			= Settings.getInstance();
	private static final EventBus	eventBus			= EventBus.getInstance();
	private static final Logger		logger				= EnvoyLog.getLogger(ChatScene.class);
	private static final int		MAX_MESSAGE_LENGTH	= 255;

	@FXML
	private void initialize() {

		// Initialize message and user rendering
		messageList.setCellFactory(listView -> new MessageListCell());
		userList.setCellFactory(listView -> new ContactListCell());

		// Listen to received messages
		eventBus.register(MessageCreationEvent.class, e -> {
			final var	message	= e.get();
			final var	chat	= localDB.getChats().stream().filter(c -> c.getRecipient().getID() == message.getSenderID()).findAny().get();

			// Update UI if in current chat
			if (chat == currentChat) Platform.runLater(() -> messageList.getItems().add(message));
		});

		// Listen to message status changes
		eventBus.register(MessageStatusChangeEvent.class, e -> {
			final var message = localDB.getMessage(e.getID());
			message.setStatus(e.get());

			// Update UI if in current chat
			if (currentChat != null && message.getSenderID() == currentChat.getRecipient().getID()) Platform.runLater(messageList::refresh);
		});

		// Listen to user status changes
		eventBus.register(UserStatusChangeEvent.class, e -> Platform.runLater(userList::refresh));

		// Listen to contacts changes
		eventBus.register(ContactOperationEvent.class, e -> {
			final var contact = e.get();
			switch (e.getOperationType()) {
				case ADD:
					localDB.getUsers().put(contact.getName(), contact);
					localDB.getChats().add(new Chat(contact));
					Platform.runLater(() -> userList.getItems().add(contact));
					break;
				case REMOVE:
					localDB.getUsers().remove(contact.getName());
					localDB.getChats().removeIf(c -> c.getRecipient().getID() == contact.getID());
					Platform.runLater(() -> userList.getItems().removeIf(c -> c.getID() == contact.getID()));
					break;
			}
		});
	}

	/**
	 * @param sceneContext the scene context used to load other scenes
	 * @param localDB      the local database form which chats and users are loaded
	 * @param client       the client used to request ID generators
	 * @param writeProxy   the write proxy used to send messages and other data to
	 *                     the server
	 * @since Envoy Client v0.1-beta
	 */
	public void initializeData(SceneContext sceneContext, LocalDB localDB, Client client, WriteProxy writeProxy) {
		this.sceneContext	= sceneContext;
		this.localDB		= localDB;
		this.client			= client;
		this.writeProxy		= writeProxy;

		userList.setItems(FXCollections.observableList(localDB.getChats().stream().map(Chat::getRecipient).collect(Collectors.toList())));
	}

	@FXML
	private void userListClicked() {
		final Contact user = userList.getSelectionModel().getSelectedItem();
		if (user != null && (currentChat == null || user.getID() != currentChat.getRecipient().getID())) {
			contactLabel.setText(user.getName());

			// LEON: JFC <===> JAVA FRIED CHICKEN <=/=> Java Foundation Classes

			// Load the chat or create a new one and add it to the LocalDB
			currentChat = localDB.getChats()
				.stream()
				.filter(c -> c.getRecipient().getID() == user.getID())
				.findAny()
				.orElseGet(() -> { final var chat = new Chat(user); localDB.getChats().add(chat); return chat; });

			messageList.setItems(FXCollections.observableList(currentChat.getMessages()));

			remainingChars.setVisible(true);
			remainingChars
				.setText(String.format("remaining chars: %d/%d", MAX_MESSAGE_LENGTH - messageTextArea.getText().length(), MAX_MESSAGE_LENGTH));
		}
		messageTextArea.setDisable(currentChat == null);
	}

	@FXML
	private void postButtonClicked() { postMessage(); }

	@FXML
	private void settingsButtonClicked() {
		sceneContext.load(SceneContext.SceneInfo.SETTINGS_SCENE);
		sceneContext.<SettingsScene>getController().initializeData(sceneContext);
	}

	@FXML
	private void addContactButtonClicked() {
		sceneContext.load(SceneContext.SceneInfo.CONTACT_SEARCH_SCENE);
		sceneContext.<ContactSearchScene>getController().initializeData(sceneContext);
	}

	@FXML
	private void messageTextUpdated(KeyEvent e) {
		// TODO: Letting go of ctrl automatically posts a message. Needs to be fixed
		// soon.

		// Automatic sending of messages via (ctrl +) enter
		if (!postButton.isDisabled() && settings.isEnterToSend() && e.getCode() == KeyCode.ENTER
				|| !settings.isEnterToSend() && e.getCode() == KeyCode.CONTROL)
			postMessage();
		else postButton.setDisable(messageTextArea.getText().isBlank() || currentChat == null);

		// Truncating messages that are too long and staying at the same position
		if (messageTextArea.getText().length() >= MAX_MESSAGE_LENGTH) {
			messageTextArea.setText(messageTextArea.getText().substring(0, MAX_MESSAGE_LENGTH));
			messageTextArea.positionCaret(MAX_MESSAGE_LENGTH);
			messageTextArea.setScrollTop(Double.MAX_VALUE);
		}

		// Redesigning the remainingChars - Label
		final int	currentLength	= messageTextArea.getText().length();
		final int	remainingLength	= MAX_MESSAGE_LENGTH - currentLength;
		remainingChars.setText(String.format("remaining chars: %d/%d", remainingLength, MAX_MESSAGE_LENGTH));
		remainingChars.setTextFill(Color.rgb(currentLength, remainingLength, 0, 1));
	}

	/**
	 * Sends a new message to the server based on the text entered in the
	 * messageTextArea.
	 *
	 * @since Envoy Client v0.1-beta
	 */
	private void postMessage() {

		// Create and send message
		sendMessage(new MessageBuilder(localDB.getUser().getID(), currentChat.getRecipient().getID(), localDB.getIDGenerator())
			.setText(messageTextArea.getText().strip())
			.build());

		// Clear text field and disable post button
		messageTextArea.setText("");
		postButton.setDisable(true);
	}

	/**
	 * Sends a message to the server and appends it to the current chat. If all
	 * message IDs have been used, a new ID generator is requested.
	 *
	 * @param message the message to send
	 * @since Envoy Client v0.1-beta
	 */
	private void sendMessage(Message message) {
		try {

			// Send message
			writeProxy.writeMessage(message);

			// Add message to LocalDB and update UI
			messageList.getItems().add(message);

			// Request a new ID generator if all IDs were used
			if (!localDB.getIDGenerator().hasNext() && client.isOnline()) client.requestIdGenerator();

		} catch (final IOException e) {
			logger.log(Level.SEVERE, "Error sending message", e);
		}
	}
}
